# -*- coding: utf-8 -*-

"""
Smolder Smoke Testing Library and cli

:copyright: (c) 2015 by Maxwell Cameron.
:license: BSD 3, see LICENSE for more details.

"""

__title__ = 'smolder'
# __version__ = '0.0.1'
# __build__ = 0x020501
__author__ = 'Max Cameron'
__license__ = 'BSD License'
__copyright__ = 'Copyright 2015 Maxwell Cameron'

from .tcptest import tcp_test
from .colours import Colours
from .colours import COLOURS
from .charcoal import Charcoal
from .get_verify import get_verify
from .get_host_overrides import get_host_overrides
from .output import Output

__all__ = ['.']



#! /usr/bin/env python
import os
import time
import logging
import warnings
from copy import deepcopy

import jsonpickle
import requests
import validictory
from yapsy.PluginManager import PluginManager

from . import COLOURS, get_verify, get_host_overrides, tcptest
from .output import Output

FORMAT = '%(asctime)-15s %(name)s [%(levelname)s]: %(message)s'
THIS_DIR = os.path.dirname(os.path.realpath(__file__))

logging.basicConfig(format=FORMAT, level=logging.ERROR, datefmt="%Y-%m-%d %H:%M:%S")
LOG = logging.getLogger('smolder')
REQUESTS_LOG = logging.getLogger('requests')
REQUESTS_LOG.setLevel(logging.ERROR)

logging.getLogger('yapsy').setLevel(logging.INFO)
manager = PluginManager()
manager.setPluginPlaces([THIS_DIR, "~/.smolder_plugins"])
manager.collectPlugins()

OUTPUT_WIDTH = 108

SCHEMA = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string"
        },
        "uri": {
            "type": "string",
            "required": False
        },
        "port": {
            "type": "integer",
            "required": False
        },
        "inputs": {
            "type": "object",
            "required": False,
            "properties": {
                "headers": {"type": "any", "required": False},
                "username": {"type": "any", "required": False},
                "password": {"type": "any", "required": False},
                "cookie": {"type": "any", "required": False},
                "data": {"type": "any", "required": False},
                "file": {"type": "any", "required": False},
                "verify": {"type": "any", "required": False},
                "allow_redirects": {"type": "any", "required": False},
                "timeout": {"type": "any", "required": False},
                "proxies": {"type": "any", "required": False}
            }
        },
        "outcomes": {
            "type": "object",
            "required": False
        },
        "protocol": {
            "type": "string",
            "required": False,
            "enum": ["tcp", "http", "https"]
        },
        "method": {
            "type": "string",
            "required": False,
            "enum": ["GET", "get", "post", "POST", "put", "PUT", "delete", "DELETE", "option", "OPTION"]
        },
        "request_headers": {
            "type": "None",
            "required": False
        },
        "url": {
            "type": "None",
            "required": False
        }
    }
}


def deepupdate(original, update):
    """
    Recursively update a dict.
    Subdict's won't be overwritten but also updated.
    :param update:
    :param original:
    """
    for key, value in original.items():
        if key not in update:
            update[key] = value
        elif isinstance(value, dict):
            deepupdate(value, update[key])
    return update


class Charcoal(object):
    def __init__(self, test, host, output_format=None):
        """

        :rtype : object
        """

        self.passed = 0
        self.failed = 0
        self.duration_ms = 0
        try:
            validictory.validate(test, SCHEMA)
            LOG.debug("Valid schema")
        except ValueError as error:
            LOG.error("Error, invalid test format: {0}.  Tests now use v0.2 format. v0.1 branch is still available."
                      .format(error))
            raise

        try:
            self.port = test["port"]
        except (AttributeError, KeyError):
            print("Warning: No port definition found in the first test, using port 80 as default.")
            try:
                if test["protocol"] == "https":
                    self.port = 443
                else:
                    self.port = 80
            except (AttributeError, KeyError):
                self.port = 80

        self.output = Output(output_format=output_format)
        LOG.debug("Test: {0}".format(test))
        test_defaults = dict(inputs=dict(allow_redirects=False, timeout=30),
                             method="get",
                             outcomes=dict(expect_status_code=200, colour_output=True))

        host_overrides = get_host_overrides.get_host_overrides(host, self.port)

        if host_overrides['hostname'] is not None:
            self.host = host_overrides['hostname']
        else:
            self.host = host
        intermediate_dict = deepupdate(test_defaults, host_overrides)

        final_dict = deepupdate(intermediate_dict, test)

        if "tcp_test" in test and test["tcp_test"]:
            tcptest.tcp_test(self.host, self.port)

        try:
            verify = final_dict["inputs"]["verify"]
        except (AttributeError, KeyError):
            verify = None
        try:
            proto = final_dict['protocol']
        except (AttributeError, KeyError):
            proto = None

        (self.verify, self.verify_specified) = get_verify.get_verify(verify, proto)

        self.test = deepcopy(final_dict)
        LOG.debug("Test with defaults: {0}".format(self.test))
        if "verify" in self.test["inputs"]:
            del self.test["inputs"]["verify"]
        self.inputs = deepcopy(self.test['inputs'])
        request_url_format = '{protocol}://{host}:{port}{uri}'
        try:
            self.inputs['url'] = request_url_format.format(protocol=self.test['protocol'], host=self.host,
                                                           port=self.test['port'],
                                                           uri=self.test['uri'])
        except KeyError:
            self.inputs['url'] = request_url_format.format(protocol=self.test['protocol'], host=self.host,
                                                           port=self.test['port'],
                                                           uri='')
        LOG.debug("Testing {0}".format(self.inputs['url']))
        self.output.append("-" * OUTPUT_WIDTH)

        this_name = ("{0:^" + str(OUTPUT_WIDTH) + "}").format(self.test['name'][:OUTPUT_WIDTH])
        self.output.append(this_name)
        this_url = ("{0:^" + str(OUTPUT_WIDTH) + "}").format(self.inputs["url"][:OUTPUT_WIDTH])
        self.output.append(this_url)
        self.output.append(self.__repr__())
        self.output.append("-" * OUTPUT_WIDTH)
        self.output.append(self.inputs, sec='request_inputs')
        self.run()

    def run(self):
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            try:
                LOG.debug("Verify is a: {0}, with value: {1}".format(type(self.verify), self.verify))
                start = int(round(time.time() * 1000))
                if self.test["protocol"] != 'tcp':
                    self.req = getattr(requests, self.test['method'].lower())(verify=self.verify, **self.inputs)
                else:
                    tcptest.tcp_test(self.host, self.port)
                end = int(round(time.time() * 1000))
                self.duration_ms = end - start
            except (RuntimeWarning, requests.exceptions.SSLError):
                warnings.simplefilter("ignore")
                start = int(round(time.time() * 1000))
                try:
                    self.req = getattr(requests, self.test['method'].lower())(verify=self.verify, **self.inputs)
                except requests.exceptions.SSLError:
                    message, status = self.fail_test("Certificate verify failed and not ignored by inputs['verify']")
                    self.add_output("SSLVerify", message, status)
                    return
                end = int(round(time.time() * 1000))
                self.duration_ms = end - start
                if not self.verify_specified:
                    message, status = self.fail_test("Insecure request not ignored by inputs['verify']")
                    self.add_output("SecureRequest", message, status)
                else:
                    if self.verify:
                        message, status = self.fail_test("Insecure request made")
                        self.add_output("SecureRequest", message, status)
                    else:
                        message, status = self.warn_test("Insecure request made and ignored")
                        self.add_output("SecureRequest", message, status)
        if self.test["protocol"] != 'tcp':
            self.output.append(dict(self.req.headers), sec='response_headers')
            self.output.append(self.req.status_code, sec='response_status_code')
            if 'show_body' in self.test:
                try:
                    req_content = self.req.content.decode()
                except UnicodeDecodeError:
                    req_content = self.req.content
                self.output.append(req_content)

            for plugin_info in manager.getAllPlugins():
                for outcome in self.test['outcomes']:
                    if plugin_info.name == outcome:
                        manager.activatePluginByName(plugin_info.name)
                        message, status = plugin_info.plugin_object.run(self)
                        self.add_output(plugin_info.name, message, status)
                        manager.deactivatePluginByName(plugin_info.name)

    def __str__(self):
        return str(self.output)

    def __repr__(self):
        LOG.debug(self.inputs)
        if "headers" in self.inputs:
            if self.inputs["headers"] != '':
                header_list = ['"{0}: {1}"'.format(k, v) for k, v in list(self.inputs["headers"].items())]
                header = "-H " + (" -H ".join(header_list))
            else:
                header = self.inputs["headers"]
        else:
            header = ''
        if "data" in self.inputs:
            data = '-d' + str(self.inputs["data"])
        else:
            data = ''

        if not self.verify and self.test["protocol"] == "https":
            curl_insecure = '--insecure'
        else:
            curl_insecure = ''

        # Adding curl output to allow simple debugging of the requests
        command = 'curl {curl_insecure} -v -s -o /dev/null {headers} {data} -X {method} "{uri}"'
        output = command.format(method=str(self.test['method']).upper(), headers=header, data=data,
                                uri=self.inputs['url'], curl_insecure=curl_insecure)
        return output

    def pass_test(self, message):
        self.passed += 1
        status = "[PASS]"
        if self.test["outcomes"]["colour_output"]:
            status = COLOURS.to_green(status)
        return message, status

    def fail_test(self, message):
        self.failed += 1
        status = "[FAIL]"
        if self.test["outcomes"]["colour_output"]:
            status = COLOURS.to_red(status)
        return message, status

    def warn_test(self, message):
        self.passed += 1
        status = "[WARN]"
        if self.test["outcomes"]["colour_output"]:
            status = COLOURS.to_yellow(status)
        return message, status

    def add_output(self, name, message, status):
        test_out = name + ": " + message + "." * (OUTPUT_WIDTH - len(name) - len(message) - 8) + status.rjust(8)
        self.output.append(test_out)

    def _to_json(self):
        return jsonpickle.encode(self)



#!/usr/bin/env python
import json
from collections import defaultdict


class Output(defaultdict):

    def __init__(self, output_format=None):
        super(Output, self).__init__(msg='')
        self.format = output_format

    def __str__(self):
        if self.format == 'json':
            ret = json.dumps(self)
        else:
            ret = self['msg']
        return ret

    def __repr__(self):
        return str(self)

    def append(self, value, sec='msg'):
        if sec == 'msg':
            self['msg'] = self['msg'] + value + '\n'
        else:
            self[sec] = value



from yapsy.IPlugin import IPlugin
import logging
LOG = logging.getLogger('smolder')


class ResponseBodyContains(IPlugin):

    @staticmethod
    def run(req):
        # Did we expect something specific in the response body?
        required_text = req.test['outcomes']['response_body_contains']
        LOG.debug(required_text)
        try:
            req_content = req.req.content.decode()
        except UnicodeDecodeError:
            req_content = req.req.content
        LOG.debug("Content: {0}".format(req_content))
        if required_text not in req_content:
            return req.fail_test("\"{0}\" does not contain \"{1}\"".format(req_content, required_text))
        else:
            return req.pass_test("Body contains \"{0}\"".format(required_text))



from yapsy.IPlugin import IPlugin
import logging
LOG = logging.getLogger('smolder')


class ResponseBodyDoesntContain(IPlugin):

    @staticmethod
    def run(req):

        # Do we need to ensure something does NOT appear in the response body?
        banned_text = req.test['outcomes']['response_body_doesnt_contain']
        try:
            req_content = req.req.content.decode()
        except UnicodeDecodeError:
            req_content = req.req.content
        if banned_text in req_content:
            return req.fail_test("Body contains \"{0}\" and shouldn't".format(banned_text))
        else:
            return req.pass_test("Body doesn't contain \"{0}\"".format(banned_text))



from yapsy.IPlugin import IPlugin
import logging
LOG = logging.getLogger('smolder')


class ResponseMaxTimeMs(IPlugin):

    @staticmethod
    def run(req):

        # Check response time
        if req.duration_ms > int(req.test['outcomes']['response_max_time_ms']):
            longer_by_ms = req.duration_ms - int(req.test['outcomes']['response_max_time_ms'])
            return req.fail_test('Response time was {0}ms longer than {1}ms max ({2}ms)'.format(str(longer_by_ms), req.test['outcomes']['response_max_time_ms'], str(req.duration_ms)))
        else:
            shorter_by_ms = int(req.test['outcomes']['response_max_time_ms']) - req.duration_ms
            return req.pass_test('Response time was {0}ms shorter than max ({1}ms)'.format(str(shorter_by_ms), str(req.duration_ms)))



from yapsy.IPlugin import IPlugin
import dpath
import logging
LOG = logging.getLogger('smolder')


class ResponseJsonContains(IPlugin):

    @staticmethod
    def run(req):
        # Validate presence of partial dicts in response json
        for path in list(req.test['outcomes']['response_json_contains'].keys()):
            expected_value = req.test['outcomes']['response_json_contains'][path]
            actual_value = dpath.util.search(req.req.json(), path)[path]
            if expected_value == actual_value:
                return req.pass_test("Body contains expected json value at path \"{0}\"".format(path))
            else:
                return req.fail_test("Invalid json value {0} at path \"{1}\", expecting {2}".format(actual_value, path, expected_value))



#!/usr/bin/env python
import logging
import socket
from retrying import retry
from .colours import COLOURS


FORMAT = '%(asctime)-15s %(name)s [%(levelname)s]: %(message)s'
logging.basicConfig(format=FORMAT, level=logging.ERROR, datefmt="%Y-%m-%d %H:%M:%S")
LOG = logging.getLogger('smolder')


@retry(wait_exponential_multiplier=500, wait_exponential_max=30000, stop_max_attempt_number=7)
def tcp_test(host, port):
    """
    Attempts to make a TCP socket connection on the specified host and
    port. Returns true if successful. Else returns false.

    """
    LOG.debug("TCP test called")
    my_sock = None
    try:
        my_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        my_sock.settimeout(1)
        my_sock.connect((host, port))
        print("Connecting to {0} on port {1} {2}".format(host, port, COLOURS.to_green("[PASS]")))
        my_sock.shutdown(2)
        my_sock.close()
    except socket.error:
        my_sock.close()
        print("Connecting to {0} on port {1} {2}".format(host, port, COLOURS.to_yellow("[FAIL]")))
        LOG.debug("Waiting for {0}:{1} to accept a connection".format(host, port))
        raise
    except Exception as error:
        print("Connecting to {0} on port {1} {2}".format(host, port, COLOURS.to_yellow("[FAIL]")))
        LOG.debug("TCP test failed: {0}".format(error))
        raise



from yapsy.IPlugin import IPlugin
import logging
LOG = logging.getLogger('smolder')


class HeadersPresent(IPlugin):

    @staticmethod
    def run(req):
        for header in req.test['response_headers_present']:
            if header not in req.req.headers:
                return req.fail_test("Expected header {0} not present".format(header))
            else:
                return req.pass_test("Header {0}: {1} present".format(header, req.req.headers[header]))




from yapsy.IPlugin import IPlugin


class ExpectStatusCode(IPlugin):

    @staticmethod
    def run(req):
        if 'expect_status_code' in req.test['outcomes']:
            if int(req.test['outcomes']['expect_status_code']) == req.req.status_code:
                return req.pass_test("Status code == {0}".format(req.test['outcomes']['expect_status_code']))
            else:
                return req.fail_test("Expecting status code {0} but got {1} instead".format(req.test['outcomes']['expect_status_code'], str(req.req.status_code)))
        else:
            if req.req.status_code == 200:
                return req.pass_test("Status code == 200")
            else:
                return req.fail_test("Status code != 200: {0}".format(req.req.status_code))



from yapsy.IPlugin import IPlugin
import logging
import re
LOG = logging.getLogger('smolder')


class ResponseHeaderValueContains(IPlugin):

    @staticmethod
    def run(req):

        for header in req.test['outcomes']['response_header_value_contains']:
            if header not in req.req.headers:
                return req.fail_test("Expected header {0} not present".format(header))
            elif not re.search(req.test['outcomes']['response_header_value_contains'][header], req.req.headers[header]):
                return req.fail_test("Expecting {0}: {1}, got {2}: {3}".format(
                    header,
                    req.test['outcomes']['response_header_value_contains'][header],
                    header,
                    req.req.headers[header]))
            else:
                return req.pass_test("Header {0}: {1} present".format(header, req.req.headers[header]))



#!/usr/bin/env python

import os
import sys

from setuptools import setup

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

PACKAGES = [
    'smolder',
]

REQUIRES = [
    'argh >= 0.26.1',
    'decorator >= 3.4.0',
    'dpath >= 1.2-70',
    'ecdsa >= 0.11',
    'Jinja2 >= 2.7.3',
    'MarkupSafe >= 0.23',
    'requests >= 2.5.0',
    'retry >= 0.4.2',
    'retrying >= 1.3.3',
    'six >= 1.8.0',
    'jsonpickle >= 0.9.2',
    'PyYAML >= 3.11',
    'Yapsy >= 1.11.23',
    'validictory >= 1.0.0',
    'nose >= 1.3.7',
    'httpretty == 0.8.10',
]

with open('README.md', 'r') as f:
    README = f.read()

setup(
    name='smolder',
    version='0.5.8',
    description='Json wrapper around requests for simple smoke testing.',
    long_description=README,
    author='Max Cameron',
    author_email='maxwell.cameron@johngaltsystems.com',
    url='http://sky-shiny.github.io/smolder/',
    packages=PACKAGES,
    package_data={'': ['LICENSE', '*.yapsy-plugin']},
    package_dir={'smolder': 'charcoal'},
    scripts=['smolder'],
    include_package_data=True,
    install_requires=REQUIRES,
    license='BSD License',
    zip_safe=False,
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP :: Site Management :: Link Checking',
        'Topic :: Internet :: WWW/HTTP :: Site Management',
        'Topic :: Internet :: WWW/HTTP :: Session',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Security',
        'Topic :: Software Development',
        'Topic :: Software Development :: Internationalization',
        'Topic :: Software Development :: Testing'
    ), requires=['requests', 'httpretty', 'nose', 'jsonpickle', 'argh', 'decorator', 'dpath',
                 'ecdsa', 'Jinja2', 'MarkupSafe', 'requests', 'retry', 'retrying', 'six',
                 'jsonpickle', 'PyYAML', 'Yapsy', 'validictory', 'nose', 'httpretty']
)



from yapsy.IPlugin import IPlugin
import validictory


class ValidateJson(IPlugin):

    @staticmethod
    def run(req):
        try:
            response_json = req.req.json()
            validictory.validate(response_json, req.test['outcomes']['validate_json']['schema'])
            return req.pass_test("Json response matches the schema provided")
        except ValueError as error:
            return req.fail_test("Json response fails to match the schema provided {0}".format(error))



#!/usr/bin/env python

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


def get_host_overrides(host, port):
    """

    :param port:
    :type host: str
    :param host:
    """
    host_overrides_object = urlparse(host)
    host_overrides = dict()
    if host_overrides_object.scheme is not None and host_overrides_object.scheme is not "":
        host_overrides["protocol"] = host_overrides_object.scheme
    else:
        host_overrides["protocol"] = 'http'
    if host_overrides_object.port is not None:
        host_overrides["port"] = host_overrides_object.port
    else:
        host_overrides["port"] = port
    host_overrides['hostname'] = host_overrides_object.hostname
    return host_overrides



#!/usr/bin/env python
def get_verify(verify=None, protocol=None):
    """

    :rtype: (bool, bool)
    :return: Should we verify the ssl request and did the user request that we do so.
    :param verify: bool | str
        The argument to pass through to requests to verify SSL.
    :param protocol: str
        Default protocol is https.  Is used to provide defaults for verify.
    """
    if verify is not None:
        if type(verify) == bool:
            intermediate_verify = verify
        elif verify == "True":
            intermediate_verify = True
        elif verify == "False":
            intermediate_verify = False
        elif type(verify) == str:
            intermediate_verify = verify
        else:
            raise TypeError('Should never happen')
        return intermediate_verify, True
    else:
        if 'https' in protocol:
            return True, False
        else:
            return False, False



from yapsy.IPlugin import IPlugin
import logging
import re
LOG = logging.getLogger('smolder')


class ResponseRedirect(IPlugin):

    @staticmethod
    def run(req):
        if "30" in str(req.req.status_code):
            match = re.match(req.test['outcomes']['response_redirect'], req.req.headers['location'])
            if match:
                message = req.pass_test("Redirect to {0}".format(req.test['outcomes']['response_redirect']))
                return message
            else:
                return req.fail_test("Got redirected to {0} instead of {1}".format(
                    req.req.headers['location'],
                    req.test['outcomes']['response_redirect']))
        else:
            return req.fail_test("Not being redirected: {0}".format(req.req.status_code))



#! /usr/bin/env python
class Colours(dict):
    def __init__(self):
        dict.__init__(dict(self))
        self.yellow = '\033[93m'
        self.green = '\033[92m'
        self.red = '\033[91m'
        self.reset = '\033[0m'

    def to_yellow(self, message):
        return self.yellow + message + self.reset

    def to_green(self, message):
        return self.green + message + self.reset

    def to_red(self, message):
        return self.red + message + self.reset

COLOURS = Colours()
